/*
 * udev.c
 *
 * Userspace devfs
 *
 * Copyright (C) 2003 Greg Kroah-Hartman <greg@kroah.com>
 * Portions Copyright (C) 2003 MontaVista Software <sdake@mvista.com>
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 * 
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 * 
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <linux/sdeq.h>

#include "udev.h"
#include "udev_version.h"
#include "sdequser.h"


#ifndef SDEQ
static char *get_action(void)
{
	char *action;

	action = getenv("ACTION");
	return action;
}


static char *get_device(void)
{
	static char device[255];
	char *temp;

	temp = getenv("DEVPATH");
	if (temp == NULL)
		return NULL;
	strcpy(device, SYSFS_ROOT);
	strcat(device, temp);

	return device;
}
#endif

/* 
 * Right now the major/minor of a device is stored in a file called
 * "dev" in sysfs.
 * The number is stored as:
 * 	MMmm
 * 		MM is the major
 * 		mm is the minor
 * 		The value is in hex.
 * Yes, this will probably change when we go to a bigger major/minor
 * range, and will have to be changed at that time.
 */
static int get_major_minor (char *dev, int *major, int *minor)
{
	char filename[255];
	char line[20];
	char temp[3];
	int fd;
	int retval = 0;

	/* add the dev file to the directory and see if it's present */
	strncpy(filename, dev, sizeof(filename));
	strncat(filename, DEV_FILE, sizeof(filename));
	fd = open(filename, O_RDONLY);
	if (fd < 0) {
		dbg("Can't open %s", filename);
		return -ENODEV;
	}

	/* get the major/minor */
	retval = read(fd, line, sizeof(line));
	if (retval < 0) {
		dbg("read error on %s", dev);
		goto exit;
	}

	temp[0] = line[0];
	temp[1] = line[1];
	temp[2] = 0x00;
	*major = (int)strtol(&temp[0], NULL, 16);

	temp[0] = line[2];
	temp[1] = line[3];
	temp[2] = 0x00;
	*minor = (int)strtol(&temp[0], NULL, 16);

	dbg("found major = %d, minor = %d", *major, *minor);

	retval = 0;
exit:
	close(fd);
	return retval;
}

/*
 * Here would go a call to the naming deamon, to get the name we want to have
 * for this device.  But for now, let's just default to whatever the kernel is
 * calling the device as that will keep the "old-style" naming policy
 */
static char *get_name(char *dev, int major, int minor)
{
	static char name[100];
	char *temp;

	temp = strrchr(dev, '/');
	if (temp == NULL)
		return NULL;
	strncpy(name, &temp[1], sizeof(name));

	dbg("name is %s", name);

	return &name[0];
}

/*
 * Again, this will live in the naming deamon
 */
static int get_mode(char *name, char *dev, int major, int minor)
{
	/* just default everyone to rw for the world! */
	return 0600;
}

/*
 * previous fork could create device nodes out of order
 *	use syscall instead
 */
static int create_node(char *name, int s_if_type, int major, int minor, int mode)
{
	char filename[255];
	dev_t dev;
	int result;

	strncpy(filename, UDEV_ROOT, sizeof(filename));
	strncat(filename, name, sizeof(filename));

	mode |= s_if_type;
	dev = ((major << 8) + minor);
	result = mknod (filename, mode, dev);

	return (result);
}

/*
 * We also want to clean up any symlinks that were created in create_node()
 */
static int delete_node(char *name)
{
	char filename[255];

	strncpy(filename, UDEV_ROOT, sizeof(filename));
	strncat(filename, name, sizeof(filename));

	dbg("unlinking %s", filename);
	return unlink(filename);
}

static int add_device(char *device, int s_if_type)
{
	char *name;
	int major;
	int minor;
	int mode;
	int retval = -EINVAL;

	retval = get_major_minor(device, &major, &minor);
	if (retval) {
		dbg ("get_major_minor failed");
		goto exit;
	}

	name = get_name(device, major, minor);
	if (name == NULL) {
		dbg ("get_name failed");
		retval = -ENODEV;
		goto exit;
	}

	mode = get_mode(name, device, major, minor);
	if (mode < 0) {
		dbg ("get_mode failed");
		retval = -EINVAL;
		goto exit;
	}

	return create_node(name, s_if_type, major, minor, mode);

exit:
	return retval;
}

static int remove_device(char *device)
{
	char *name;
	int retval = 0;

	name = get_name(device, 0, 0);
	if (name == NULL) {
		dbg ("get_name failed");
		retval = -ENODEV;
		goto exit;
	}

	return delete_node(name);

exit:
	return retval;
}

int process_event (struct sdeq_event *event)
{
	int s_if_type;
	char path[255];

	if (strcmp(event->device_type, "block") == 0)
		s_if_type = S_IFBLK;
	else
		s_if_type = S_IFCHR;


	sprintf (path, "%s%s", SYSFS_ROOT, event->kobj_path);
	if (event->event_type == SDEQ_EVENT_TYPE_INSERT) {
		return add_device(path, s_if_type);
	} else 
	if (event->event_type == SDEQ_EVENT_TYPE_REMOVE) {
		return remove_device(path);
	}

	return 0;
}

int main (void)
{
	struct sdeq_event event;
	int result;
	int fd;
	fd_set read_fds;

#ifndef DEBUG
	switch (fork ()) {
	case -1:
		exit (1);
		break;
	case 0:
		/*
		 * child which is disconnected, run this process
		 */
		break;
	default:
		exit (0);
		break;
	}
#endif

	sdeq_init (&fd);
	sdeq_reset_currptr ();
	for (;;) {
		FD_ZERO (&read_fds);
		FD_SET (fd, &read_fds);

retry_select:
		result = select (fd + 1, &read_fds, NULL, NULL, NULL);
		printf ("done with select\n");
		if (result == -1 && errno == EINTR) {
			goto retry_select;
		} else
		if (result == -1) {
			perror ("select: ");
			exit (1);
		}

		/*
		 * Process all events
		 * */
		do {
			result = sdeq_get_currevent (&event);
			if (result == 0) {
				process_event (&event);
				sdeq_remove_events (1);	
			}
		} while (result == 0);
	}
}

